module M where

-- Monads with operations for exceptions and counter
--
-- These operations will be implemented for the plain
-- state and exception monads but also for the transformed
-- variants.

class Monad m => ExMonad m where
  raise :: String -> m a

class Monad m => CtMonad m where
  put :: Int -> m ()
  get :: m Int

-- Plain exception monad
--
-- Note: we do not use Maybe but attach strings as exceptions.

data Exception a = Val a | Exn String

instance Monad Exception where
  return x   = Val x
  m >>= f    = case m of Exn x -> Exn x ; Val x -> f x

instance Functor Exception where
  fmap f m = m >>= \x -> return (f x)
instance Applicative Exception where
  pure = return
  m1 <*> m2 = m1 >>= (\x1 -> m2 >>= (\x2 -> return (x1 x2)))

instance ExMonad Exception where
  raise x = Exn x

-- Plain counter monad

data Counter a = Cntr (Int -> (a,Int))

instance Monad Counter where
  return x = Cntr (\i -> (x,i))
  Cntr f >>= g = Cntr (\i -> let (v,j) = f i in let Cntr gg = g v in gg j)

instance Functor Counter where
  fmap f m = m >>= \x -> return (f x)
instance Applicative Counter where
  pure = return
  m1 <*> m2 = m1 >>= (\x1 -> m2 >>= (\x2 -> return (x1 x2)))

instance CtMonad Counter where
  put i = Cntr (\_ -> ((),i))
  get = Cntr (\s -> (s,s))

--- Monad transformers

class MonadTrans t where
  lift :: Monad m => m a -> t m a
 
--- Exception transformer

data ExceptionT m a = MkExc (m (Exception a))

instance Monad m => Monad (ExceptionT m) where
  return x = MkExc (return (Val x))
  (MkExc m) >>= n = MkExc (m >>= r) where r (Exn y) = return (Exn y)
                                          r (Val y) = let MkExc r = n y in r

-- Please ignore this... {{{
instance Monad m => Functor (ExceptionT m) where
  fmap f m = m >>= \x -> return (f x)
instance Monad m => Applicative (ExceptionT m) where
  pure = return
  m1 <*> m2 = m1 >>= (\x1 -> m2 >>= (\x2 -> return (x1 x2)))
-- }}}

instance Monad m => ExMonad (ExceptionT m) where
  raise x = MkExc (return (Exn x))

instance MonadTrans ExceptionT where
  lift g = MkExc (do x <- g; return (Val x))

-- Lifting state monad operations with the exception transformer

instance CtMonad m => CtMonad (ExceptionT m) where
  get = lift get
  put x = lift (put x)

-- We could do all these steps for a counter monad transformer too.

--- Back to our example: the evaluator

data Expr = Entier Int | Mult Expr Expr | Div Expr Expr

test :: Expr
test = Mult (Entier 3) (Div (Entier 6) (Entier 2))

incr :: ExceptionT Counter ()
incr = do x <- get ; put (x+1)

eval :: Expr -> ExceptionT Counter Int
eval (Entier i) = return i
eval (Mult e f) = do incr ; x1 <- eval e ; x2 <- eval f ; return (x1 * x2)
eval (Div e f) = do incr ; x1 <- eval e ; x2 <- eval f ;
                    if x2 == 0 then raise "div by zero" else
                      return (x1 `div` x2)
